# Generated file
# - do not overwrite
# - do not include in git
from collections import namedtuple

BuildInfo = namedtuple(
    'BuildInfo',
    ['version', 'built_at', 'sha', 'branch', 'tag'],
)

BUILD_INFO = BuildInfo(
    version='v3.19.2~d8981fa',
    built_at='2023-01-17 16:09:22Z',
    sha='d8981fa6a7bc136ca3d41b43fc768c68dedb7e96',
    branch='HEAD',
    tag='v3.19.2',
)
